@@ -20,8 +20,6 @@
 
 from __future__ import unicode_literals
 
-import json as libjson
-import pickle
 import sys
 
 from RestAuthCommon import error
@@ -30,14 +28,8 @@ PY2 = sys.version_info[0] == 2
 PY3 = sys.version_info[0] == 3
 
 if PY2:  # pragma: py2
-    from urlparse import parse_qs
-    from urllib import urlencode
-
     string_types = basestring
 else:  # pragma: py3
-    from urllib.parse import parse_qs
-    from urllib.parse import urlencode
-
     string_types = (str, bytes, )
 
 
@@ -117,7 +109,7 @@ class ContentHandler(object):
                 return self._normalize_dict3(v)
             return v
 
-        return dict((conv(k), conv(v)) for k, v in d.items())
+        return {conv(k): conv(v) for k, v in d.items()}
 
     def _normalize_dict2(self, d):  # pragma: py2
         """Converts any keys or values of d that are str to unicode."""
@@ -128,7 +120,7 @@ class ContentHandler(object):
                 return self._normalize_dict2(v)
             return v
 
-        return dict((conv(k), conv(v)) for k, v in d.iteritems())
+        return {conv(k): conv(v) for k, v in d.iteritems()}
 
     def _normalize_str3(self, s):  # pragma: py3
         """Converts byte objects to str."""
@@ -142,7 +134,7 @@ class ContentHandler(object):
         """Shortcut for marshalling just any object.
 
         .. NOTE:: If you know the type of **obj** in advance, you should use the marshal_* methods
-        directly for improved speed.
+            directly for improved speed.
 
         :param obj: The object to marshall.
         :return: The marshalled representation of the object.
@@ -233,82 +225,87 @@ class JSONContentHandler(ContentHandler):
        <http://en.wikipedia.org/wiki/JSON>`_
     """
 
+    librarypath = 'json'
     mime = 'application/json'
     """The mime-type used by this content handler is 'application/json'."""
 
     SEPARATORS = (str(','), str(':'))
 
-    class ByteEncoder(libjson.JSONEncoder):
+    def __init__(self, **kwargs):
+        super(JSONContentHandler, self).__init__(**kwargs)
+
         if PY3:  # pragma: py3
-            def decode_dict(self, d):
-                def conv(v):
-                    if isinstance(v, bytes):
-                        return v.decode('utf-8')
-                    elif isinstance(v, dict):
-                        return self.decode_dict(v)
-                    return v
+            class ByteEncoder(self.library.JSONEncoder):
+                def decode_dict(self, d):
+                    def key(v):  # keys are not handled by self.default()
+                        if isinstance(v, bytes):
+                            return v.decode('utf-8')
+                        return v
 
-                return dict((conv(k), conv(v)) for k, v in d.items())
+                    def val(v):  # handle nested dicts
+                        if isinstance(v, dict):
+                            return self.decode_dict(v)
+                        return v
 
-            def encode(self, obj):
-                if isinstance(obj, bytes):
-                    obj = obj.decode('utf-8')
-                elif isinstance(obj, dict):
-                    obj = self.decode_dict(obj)
+                    return {key(k): val(v) for k, v in d.items()}
 
-                try:
-                    return libjson.JSONEncoder.encode(self, obj)
-                except TypeError:
-                    raise
+                def encode(self, obj):
+                    if isinstance(obj, dict):
+                        obj = self.decode_dict(obj)
 
-            def default(self, obj):
-                if isinstance(obj, bytes):
-                    return obj.decode('utf-8')
-                return libjson.JSONEncoder.default(self, obj)
+                    return super(ByteEncoder, self).encode(obj)
+
+                def default(self, obj):  # for objects of unknown type (i.e. bytes)
+                    if isinstance(obj, bytes):
+                        return obj.decode('utf-8')
+                    return super(ByteEncoder, self).default(obj)
+
+            self.encoder = ByteEncoder
+        else:  # pragma: py2
+            self.encoder = self.library.JSONEncoder
 
     def unmarshal_str(self, body):
         try:
-            pure = libjson.loads(self.normalize_str(body))
+            pure = self.library.loads(self.normalize_str(body))
             if not isinstance(pure, list) or len(pure) != 1:
                 raise error.UnmarshalError("Could not parse body as string")
 
             string = pure[0]
 
-            # In python 2.7.1 (not 2.7.2) json.loads("") returns a str and
-            # not unicode.
+            # In python 2.7.1 (not 2.7.2) json.loads("") returns a str and not unicode.
             return self.normalize_str(string)
         except ValueError as e:
             raise error.UnmarshalError(e)
 
     def unmarshal_dict(self, body):
         try:
-            return libjson.loads(self.normalize_str(body))
+            return self.library.loads(self.normalize_str(body))
         except ValueError as e:
             raise error.UnmarshalError(e)
 
     def unmarshal_list(self, body):
         try:
-            return libjson.loads(self.normalize_str(body))
+            return self.library.loads(self.normalize_str(body))
         except ValueError as e:
             raise error.UnmarshalError(e)
 
     def marshal_str(self, obj):
         try:
-            dumped = libjson.dumps([obj], separators=self.SEPARATORS, cls=self.ByteEncoder)
+            dumped = self.library.dumps([obj], separators=self.SEPARATORS, cls=self.encoder)
             return dumped.encode('utf-8')
         except Exception as e:
             raise error.MarshalError(e)
 
     def marshal_list(self, obj):
         try:
-            dumped = libjson.dumps(obj, separators=self.SEPARATORS, cls=self.ByteEncoder)
+            dumped = self.library.dumps(obj, separators=self.SEPARATORS, cls=self.encoder)
             return dumped.encode('utf-8')
         except Exception as e:
             raise error.MarshalError(e)
 
     def marshal_dict(self, obj):
         try:
-            dumped = libjson.dumps(obj, separators=self.SEPARATORS, cls=self.ByteEncoder)
+            dumped = self.library.dumps(obj, separators=self.SEPARATORS, cls=self.encoder)
             return dumped.encode('utf-8')
         except Exception as e:
             raise error.MarshalError(e)
@@ -322,8 +319,8 @@ class BSONContentHandler(ContentHandler):
 
     .. seealso:: `Specification <http://bsonspec.org/>`_, `WikiPedia
        <http://en.wikipedia.org/wiki/BSON>`_,
-       `pymongo <https://pypi.python.org/pypi/pymongo>` on PyPi,
-       `bson <https://pypi.python.org/pypi/bson>` on PyPi
+       `pymongo <https://pypi.python.org/pypi/pymongo>`_ on PyPi,
+       `bson <https://pypi.python.org/pypi/bson>`_ on PyPi
     """
     mime = 'application/bson'
     """The mime-type used by this content handler is 'application/json'."""
@@ -335,19 +332,19 @@ class BSONContentHandler(ContentHandler):
 
         if hasattr(self.library, 'BSON'):  # pragma: pymongo
             self.dumps = self.library.BSON.encode
-            self.loads = self.library.BSON.decode
+            self.loads = lambda d: self.library.BSON(d).decode()
         else:  # pragma: libbson
             self.dumps = self.library.dumps
             self.loads = self.library.loads
 
     def marshal_dict(self, obj):
-        return self.dumps({'d': self.normalize_dict(obj), })
+        return self.marshal_cast(self.dumps({'d': self.normalize_dict(obj), }))
 
     def marshal_list(self, obj):
-        return self.dumps({'l': self.normalize_list(obj), })
+        return self.marshal_cast(self.dumps({'l': self.normalize_list(obj), }))
 
     def marshal_str(self, obj):
-        return self.dumps({'s': self.normalize_str(obj), })
+        return self.marshal_cast(self.dumps({'s': self.normalize_str(obj), }))
 
     def _unmarshal_dict2(self, body):  # pragma: py2
         # NOTE: We convert unicode because some old versions of RestAuthClient
@@ -379,10 +376,12 @@ class BSONContentHandler(ContentHandler):
         unmarshal_dict = _unmarshal_dict3
         unmarshal_list = _unmarshal_list3
         unmarshal_str = _unmarshal_str3
+        marshal_cast = bytes
     else:  # pragma: py2
         unmarshal_dict = _unmarshal_dict2
         unmarshal_list = _unmarshal_list2
         unmarshal_str = _unmarshal_str2
+        marshal_cast = str
 
 
 class MessagePackContentHandler(ContentHandler):
@@ -441,6 +440,19 @@ class FormContentHandler(ContentHandler):
 
     SUPPORT_NESTED_DICTS = False
 
+    def __init__(self, **kwargs):
+        super(FormContentHandler, self).__init__(**kwargs)
+
+        if PY2:  # pragma: py2
+            from urlparse import parse_qs
+            from urllib import urlencode
+        else:  # pragma: py3
+            from urllib.parse import parse_qs
+            from urllib.parse import urlencode
+
+        self.parse_qs = parse_qs
+        self.urlencode = urlencode
+
     def _decode_dict(self, d):  # pragma: py2
         decoded = {}
         for key, value in d.items():
@@ -455,49 +467,46 @@ class FormContentHandler(ContentHandler):
         return decoded
 
     def unmarshal_dict(self, body):
-        if PY3:  # pragma: py3
+        if PY3:  # pragma: no branch py3
             body = body.decode('utf-8')
 
-        parsed_dict = parse_qs(body, True)
+        parsed_dict = self.parse_qs(body, True)
         ret_dict = {}
         for key, value in parsed_dict.items():
-            if isinstance(value, list) and len(value) == 1:
-                ret_dict[key] = value[0]
-            else:
-                ret_dict[key] = value
+            ret_dict[key] = value[0]
 
-        if PY2:  # pragma: py2
+        if PY2:  # pragma: no branch py2
             ret_dict = self._decode_dict(ret_dict)
 
         return ret_dict
 
     def unmarshal_list(self, body):
-        if PY3:  # pragma: py3
+        if PY3:  # pragma: no branch py3
             body = body.decode('utf-8')
 
         if body == '':
             return []
 
-        parsed = parse_qs(body, True)['list']
+        parsed = self.parse_qs(body, True)['list']
 
-        if PY2:  # pragma: py2
+        if PY2:  # pragma: no branch py2
             parsed = [e.decode('utf-8') for e in parsed]
         return parsed
 
     def unmarshal_str(self, body):
-        if PY3:  # pragma: py3
+        if PY3:  # pragma: no branch py3
             body = body.decode('utf-8')
 
-        parsed = parse_qs(body, True)['str'][0]
+        parsed = self.parse_qs(body, True)['str'][0]
         return self.normalize_str(parsed)
 
     def marshal_str(self, obj):
         try:
             if PY2:  # pragma: py2
                 obj = obj.encode('utf-8')
-                return urlencode({'str': obj})
+                return self.urlencode({'str': obj})
             else:  # pragma: py3
-                return urlencode({'str': obj}).encode('utf-8')
+                return self.urlencode({'str': obj}).encode('utf-8')
         except Exception as e:
             raise error.MarshalError(e)
 
@@ -516,7 +525,7 @@ class FormContentHandler(ContentHandler):
 
     def marshal_dict(self, obj):
         try:
-            if PY2:  # pragma: py2
+            if PY2:  # pragma: no branch py2
                 obj = self._encode_dict(obj)
 
             # verify that no value is a dictionary, because the unmarshalling for
@@ -526,9 +535,9 @@ class FormContentHandler(ContentHandler):
                     raise error.MarshalError(
                         "FormContentHandler doesn't support nested dictionaries.")
             if PY3:  # pragma: py3
-                return urlencode(obj, doseq=True).encode('utf-8')
+                return self.urlencode(obj, doseq=True).encode('utf-8')
             else:  # pragma: py2
-                return urlencode(obj, doseq=True)
+                return self.urlencode(obj, doseq=True)
         except Exception as e:
             raise error.MarshalError(e)
 
@@ -536,9 +545,9 @@ class FormContentHandler(ContentHandler):
         try:
             if PY2:  # pragma: py2
                 obj = [e.encode('utf-8') for e in obj]
-                return urlencode({'list': obj}, doseq=True)
+                return self.urlencode({'list': obj}, doseq=True)
             else:  # pragma: py3
-                return urlencode({'list': obj}, doseq=True).encode('utf-8')
+                return self.urlencode({'list': obj}, doseq=True).encode('utf-8')
         except Exception as e:
             raise error.MarshalError(e)
 
@@ -554,41 +563,42 @@ class PickleContentHandler(ContentHandler):
     mime = 'application/pickle'
     """The mime-type used by this content handler is 'application/pickle'."""
 
+    librarypath = 'pickle'
     PROTOCOL = 2
 
     def marshal_str(self, obj):
         try:
-            return pickle.dumps(self.normalize_str(obj), protocol=self.PROTOCOL)
+            return self.library.dumps(self.normalize_str(obj), protocol=self.PROTOCOL)
         except Exception as e:
             raise error.MarshalError(str(e))
 
     def marshal_dict(self, obj):
         try:
-            return pickle.dumps(self.normalize_dict(obj), protocol=self.PROTOCOL)
+            return self.library.dumps(self.normalize_dict(obj), protocol=self.PROTOCOL)
         except Exception as e:
             raise error.MarshalError(str(e))
 
     def marshal_list(self, obj):
         try:
-            return pickle.dumps(self.normalize_list(obj), protocol=self.PROTOCOL)
+            return self.library.dumps(self.normalize_list(obj), protocol=self.PROTOCOL)
         except Exception as e:
             raise error.MarshalError(str(e))
 
     def unmarshal_str(self, data):
         try:
-            return self.normalize_str(pickle.loads(data))
+            return self.normalize_str(self.library.loads(data))
         except Exception as e:
             raise error.UnmarshalError(str(e))
 
     def unmarshal_list(self, data):
         try:
-            return self.normalize_list(pickle.loads(data))
+            return self.normalize_list(self.library.loads(data))
         except Exception as e:
             raise error.UnmarshalError(str(e))
 
     def unmarshal_dict(self, data):
         try:
-            return self.normalize_dict(pickle.loads(data))
+            return self.normalize_dict(self.library.loads(data))
         except Exception as e:
             raise error.UnmarshalError(str(e))
 
@@ -624,38 +634,88 @@ class YAMLContentHandler(ContentHandler):
     librarypath = 'yaml'
 
     def _marshal_str3(self, obj):  # pragma: py3
-        return self.library.dump(obj).encode('utf-8')
+        return self.library.dump(self.normalize_str(obj), encoding='utf-8')
+
+    def _py2_str_helper(self, s):  # pragma: py2
+        """Wrap eratic behaviour of the Python2 YAML implementation.
+
+        In Python2, the YAML implentation is unfortunately very eratic. Here is what the
+        documentation says:
+
+        * str objects are converted into !!str, !!python/str or !binary nodes depending on whether
+          the object is an ASCII, UTF-8 or binary string.
+        * unicode objects are converted into !!python/unicode or !!str nodes depending on whether the
+          object is an ASCII string or not.
+
+        This is unfortunately unreproduceable::
+
+            >>> import yaml
+            >>> yaml.dump('a')  # (a) this result is correct
+            'a\n...\n'
+            >>> yaml.dump(u'a')  # (b) unicode object returns non-portable !!python/unicode
+            "!!python/unicode 'a'\n"
+            >>> yaml.dump(u'a'.encode('utf-8'))  # same as yaml.dump('a')
+            'a\n...\n'
+            >>> yaml.dump(u'foo愑')  # (c) non-ascii unicode returns portable str
+            '"foo\\u6111"\n'
+            >>> yaml.dump(u'foo愑'.encode('utf-8'))  # (d) str returns non-portable !!python/str
+            '!!python/str "foo\\u6111"\n'
+
+        So to summarize, this method does::
+
+        * cast unicode to str if it is ASCII, so (b) is converted to (a)
+        * cast a str object to unicode if it non-ASCII: (d) is converted (c)
+        """
+        if type(s) == unicode:
+            try:
+                s.decode('utf-8')
+                return s.encode('utf-8')
+            except UnicodeEncodeError:
+                pass
+        elif type(s) == str:
+            s.encode('utf-8')
+        return s
 
     def _marshal_str2(self, obj):  # pragma: py2
-        return self.library.dump(obj)
+        return self.library.dump(self._py2_str_helper(obj))
 
     def marshal_str(self, obj):
         try:
-            return self._marshal_str(self.normalize_str(obj))
+            return self._marshal_str(obj)
         except Exception as e:
             raise error.MarshalError(e)
 
     def _marshal_dict3(self, obj):  # pragma: py3
-        return self.library.dump(obj).encode('utf-8')
+        return self.library.dump(self.normalize_dict(obj), encoding='utf-8')
+
+    def _py2_dict_helper(self, d):  # pragma: py2
+        def conv(v):
+            if isinstance(v, unicode):
+                return self._py2_str_helper(v)
+            elif isinstance(v, dict):
+                return self._py2_dict_helper(v)
+            return v
+
+        return dict((conv(k), conv(v)) for k, v in d.iteritems())
 
     def _marshal_dict2(self, obj):  # pragma: py2
-        return self.library.dump(obj)
+        return self.library.dump(self._py2_dict_helper(obj))
 
     def marshal_dict(self, obj):
         try:
-            return self._marshal_dict(self.normalize_dict(obj))
+            return self._marshal_dict(obj)
         except Exception as e:
             raise error.MarshalError(e)
 
     def _marshal_list3(self, obj):  # pragma: py3
-        return self.library.dump(obj).encode('utf-8')
+        return self.library.dump(self.normalize_list(obj), encoding='utf-8')
 
     def _marshal_list2(self, obj):  # pragma: py2
-        return self.library.dump(obj)
+        return self.library.dump([self._py2_str_helper(s) for s in obj])
 
     def marshal_list(self, obj):
         try:
-            return self._marshal_list(self.normalize_list(obj))
+            return self._marshal_list(obj)
         except Exception as e:
             raise error.MarshalError(e)
 
@@ -812,13 +872,3 @@ application/messagepack           :py:class:`.handlers.MessagePackContentHandler
 If you want to provide your own implementation of a :py:class:`.ContentHandler`, you can add it to
 this dictionary with the appropriate MIME type as the key.
 """
-
-# old names, for compatability:
-content_handler = ContentHandler
-json = JSONContentHandler
-xml = XMLContentHandler
-form = FormContentHandler
-
-# 'YamlContentHandler' was introduced in 0.6.1 and renamed for consistency to
-# 'YAMLContentHandler' in 0.6.2
-YamlContentHandler = YAMLContentHandler
